from joblib import Parallel, delayed
from oaipmh.client import Client
from pymongo import MongoClient
from datetime import datetime, timedelta
import xmltodict
import psutil


class OxomocCheckPointSelective:
    """
    Class to handle checkpoints for Colav OAI-PMH using selective strategy by dates.
    """

    def __init__(self, mongodb_uri="mongodb://localhost:27017/"):
        """
        CheckPoint constructor

        Parameters:
        ----------
        mongodb_uri:str
            MongoDB connection string uri
        """
        self.client = MongoClient(mongodb_uri)

    def create(self, base_url: str, mongo_db: str, mongo_collection: str, metadataPrefix='oai_dc', force_http_get=True, days=30, max_tries=4):  # noqa: E501
        """
        Method to create the checkpoint, this allows to save all the ids for records and sets
        in order to know what was downloaded.
        All the checkpints are saved in the mongo collections

        Parameters:
        ----------
        base_url:str
            D-Space endpoint url
        mongo_db:str
            MongoDB database name
        mongo_collection:str
            MongoDB collection name
        metadataPrefix:str
            metadata type for xml schema ex: dim, xoai, mods, oai_dc (default: oai_dc)
        force_http_get:bool
            force to use get instead post for requests
        """
        client = Client(base_url, force_http_get=force_http_get)
        try:
            identity = client.identify()
        except BaseException as err:
            print(f"=== ERROR: Unexpected {err}, {type(err)}")
            print(f"=== ERROR: CheckPoint can not be created for {base_url}")
            return
        if metadataPrefix not in [i[0] for i in client.listMetadataFormats()]:
            print(
                f"=== ERROR: metadataPrefix {metadataPrefix}, not supported for {base_url}")
            print(
                f"=== ERROR: CheckPoint can not be created for {mongo_collection} omitting..")
            return

        print(
            f"=== Creating CheckPoint for {mongo_collection} from  {base_url} with metadataPrefix {metadataPrefix}")

        info = {}
        info["repository_name"] = identity.repositoryName()
        info["admin_emails"] = identity.adminEmails()
        info["base_url"] = identity.baseURL()
        info["protocol_version"] = identity.protocolVersion()
        info["earliest_datestamp"] = identity.earliestDatestamp()
        info["granularity"] = identity.granularity()

        self.client[mongo_db][f"{mongo_collection}_identity"].drop()
        self.client[mongo_db][f"{mongo_collection}_identity"].insert_one(info)

        delta = 60 * 60 * 24
        col_identifiers = self.client[mongo_db][f"{mongo_collection}_identifiers"]
        ckp = col_identifiers.find_one({"_id": 0})
        if ckp is not None:
            if "final_date" in ckp.keys():
                init_date = ckp["final_date"]
            else:
                init_date = identity.earliestDatestamp()
        else:
            init_date = identity.earliestDatestamp()
            col_identifiers.insert_one({"_id": 0, "initial_date": init_date})
        end_date = init_date + timedelta(days=days)
        end_date = end_date.replace(hour=0, minute=0, second=0)
        if end_date > datetime.today():
            end_date = datetime.today().replace(microsecond=0)
        while init_date < datetime.today():
            print("=== INFO:", init_date, "----", end_date, mongo_collection)
            params = {"verb": "ListIdentifiers", "metadataPrefix": metadataPrefix,
                      "from": init_date.isoformat(), "until": end_date.isoformat()}
            for i in range(max_tries):
                try:
                    ids = client.makeRequest(**params)
                    break
                except BaseException as err:
                    print(f"=== ERROR: Unexpected {err}, {type(err)}")
                    print(
                        f"=== ERROR: CheckPoint try {i} of {max_tries} for {base_url}")
                    if i == (max_tries - 1):
                        print(
                            f"=== ERROR: CheckPoint can not be created for {base_url} with params {params}")
                        return

            ids = xmltodict.parse(ids)
            identifiers = []
            total = 0
            if "error" in ids['OAI-PMH'].keys():
                if ids['OAI-PMH']["error"]['@code'] == 'noRecordsMatch':
                    # records not found in the period of time
                    # setting next time range
                    col_identifiers.update_one(
                        {"_id": 0}, {"$set": {"final_date": end_date}})
                    init_date = end_date + timedelta(seconds=delta)
                    end_date = init_date + timedelta(days=days)
                    if end_date > datetime.today():
                        end_date = datetime.today().replace(microsecond=0)
                    continue
                else:
                    print("=== ERROR:", ids['OAI-PMH']["error"]
                          ['@code'], mongo_collection, base_url, params),
                    print(init_date, "----", end_date,
                          "ERROR creating checkpoint!!!", mongo_collection)
                    break
            if ids['OAI-PMH']['ListIdentifiers'] is None:
                pass
            else:
                identifiers = ids['OAI-PMH']['ListIdentifiers']["header"]
                # if there is only one register is returning a dict, instead list
                if type(identifiers) is not list:
                    identifiers = [identifiers]
                resumptionToken = "resumptionToken" in ids['OAI-PMH']['ListIdentifiers'].keys(
                )
                if not resumptionToken:
                    total = len(identifiers)
                while resumptionToken:
                    params = {}
                    params['verb'] = 'ListIdentifiers'
                    total = eval(ids['OAI-PMH']['ListIdentifiers']['resumptionToken']
                                 ['@completeListSize'])  # this is int for dspace
                    if type(total) is not int:
                        # this is for zenodo, maybe other implementations?
                        total = total["value"]
                    if '#text' in ids['OAI-PMH']['ListIdentifiers']['resumptionToken'].keys():
                        params['resumptionToken'] = ids['OAI-PMH']['ListIdentifiers']['resumptionToken']['#text']
                    else:
                        break
                    ids = client.makeRequest(**params)
                    ids = xmltodict.parse(ids)
                    _ids = ids['OAI-PMH']['ListIdentifiers']["header"]
                    # if there is only one register is returning a dict, instead list
                    if type(_ids) is not list:
                        _ids = [_ids]
                    identifiers += _ids
                    resumptionToken = "resumptionToken" in ids['OAI-PMH']['ListIdentifiers'].keys(
                    )
                    print("=== INFO:", init_date, "----", end_date, mongo_collection,
                          f"Pagination {len(identifiers)} of {total}", flush=True)
                for i in identifiers:
                    i["downloaded"] = False
            if len(identifiers) != 0:
                col_identifiers.update_one({"_id": 0}, {"$set": {"final_date": end_date},
                                                        '$push': {'identifiers': {'$each': identifiers},
                                                                  'ranges': {"init_date": init_date,
                                                                             'final_date': end_date, "n_records": total}}},
                                           upsert=True)
            else:
                col_identifiers.update_one(
                    {"_id": 0}, {"$set": {"final_date": end_date}})
            init_date = end_date + timedelta(seconds=delta)
            end_date = init_date + timedelta(days=days)
            if end_date > datetime.today():
                end_date = datetime.today().replace(microsecond=0)

    def exists_records(self, mongo_db: str, mongo_collection: str):
        """
        Method to check if the checkpoints already exists for records.

        Parameters:
        ----------
        mongo_db:str
            MongoDB database name
        mongo_collection:str
            MongoDB collection name
        """
        ckp_rec = f"{mongo_collection}_identifiers"
        collections = self.client[mongo_db].list_collection_names()
        return ckp_rec in collections

    def drop(self, mongo_db: str, mongo_collection: str):
        """
        Method to delete all the checkpoints.

        Parameters:
        ----------
        mongo_db:str
            MongoDB database name
        mongo_collection:str
            MongoDB collection name
        """
        self.client[mongo_db][f"{mongo_collection}_identity"].drop()
        self.client[mongo_db][f"{mongo_collection}_identifiers"].drop()

    def update_record(self, mongo_db: str, mongo_collection: str, keys: dict):
        """
        Method to update the status of a record in the checkpoint

        Parameters:
        ----------
        mongo_db:str
            MongoDB database name
        mongo_collection:str
            MongoDB collection name
        keys:dict
            Dictionary with _id and other required values to perform the update.
        """
        self.client[mongo_db][f"{mongo_collection}_identifiers"].update_one({"_id": 0}, {"$set": {
                                                                            "identifiers.$[idx].downloaded": True}},
                                                                            upsert=True,
                                                                            array_filters=[{'idx.identifier': keys["_id"]}])

    def get_records_regs(self, mongo_db: str, mongo_collection: str):
        """
        Function to get registers from the records ckp collection that are not downloaded

        Parameters:
        ----------
        mongo_db:str
            MongoDB database name
        mongo_collection:str
            MongoDB collection name

        Returns:
        ----------
        list
            ids of records not downloaded.
        """
        pipeline = [
            {"$match": {}},
            {"$unwind": "$identifiers"},
            {"$match": {"$and": [{"identifiers.@status": {"$ne": "deleted"}},
                                 {"identifiers.downloaded": False}]}},
            {"$addFields": {"_id": "$identifiers.identifier"}},
            {"$project": {"_id": 1}},
        ]
        ckp_col = self.client[mongo_db][f"{mongo_collection}_identifiers"]
        ckpdata = list(ckp_col.aggregate(pipeline))
        return ckpdata

    def run(self, endpoints: dict, mongo_db: str, jobs: int = None):
        """
        Method to create in parallel the checkpoints,
        every thread for endpoint

        Parameters:
        ----------
        endpoints: dict
            dictionary with the endpoints
        mongo_db: str
            database name
        jobs: int
            number of threads for the parallel execution,
            if None maximum allowed by the cpu.
        """
        if jobs is None:
            jobs = psutil.cpu_count()

        Parallel(n_jobs=jobs, backend="threading", verbose=10)(delayed(self.create)(
            endpoints[endpoint]["url"], mongo_db,
            endpoint, endpoints[endpoint]["metadataPrefix"]) for endpoint in endpoints.keys())



from oxomoc.checkpoint import OxomocCheckPoint
from oxomoc.ckpselective import OxomocCheckPointSelective
from pymongo import MongoClient
from oaipmh.client import Client
from joblib import Parallel, delayed
import psutil
import xmltodict
from ratelimit import limits, sleep_and_retry
import sys


class OxomocHarvester:
    """
    Class for harvesting data from OAI-PHM protocol
    """

    def __init__(self, endpoints: dict, mongo_db="oxomoc", mongodb_uri="mongodb://localhost:27017/", force_http_get=True):
        """
        Harvester constructor

        Parameters:
        ----------
        endpoints:dict
            dictionary with parameters for endpoints such as url, name, rate-limit, etc..
        mongo_db:str
            MongoDB name, default value "oxomoc"
        mongodb_uri:str
            MongoDB connection string uri
        """
        self.endpoints = endpoints
        self.mongo_db = mongo_db
        self.mongodb_uri = mongodb_uri
        self.client = MongoClient(mongodb_uri)
        self.force_http_get = force_http_get
        self.check_limit = {}
        self.checkpoint = {}
        for endpoint in self.endpoints.keys():
            if "rate_limit" in self.endpoints[endpoint].keys():
                calls = self.endpoints[endpoint]["rate_limit"]["calls"]
                secs = self.endpoints[endpoint]["rate_limit"]["secs"]

                @sleep_and_retry
                @limits(calls=calls, period=secs)
                def check_limit():
                    pass
                self.check_limit[endpoint] = check_limit
            else:
                def check_limit():
                    pass
                self.check_limit[endpoint] = check_limit

    def process_record(self, client: Client, identifier: str, metadataPrefix: str, endpoint: str):
        """
        This method perform the request for the given record id and save it in the mongo
        collection and updates the checkpoint collection when it was inserted.

        Parameters:
        ---------
        client: oaipmh.client
            oaipmh client instance
        identifier:str
            record id
        metadataPrefix:str
            metadata type for xml schema ex: dim, xoai, mods, oai_dc (default: oai_dc)
        endpoint:str
            name of the endpoint to process and the MongoDb collection name
        """
        self.check_limit[endpoint]()

        try:
            raw_record = client.makeRequest(
                **{'verb': 'GetRecord', 'identifier': identifier, 'metadataPrefix': metadataPrefix})
        except Exception as e:
            record = {}
            record["identifier"] = identifier
            record["instance"] = str(type(e))
            record["item_type"] = "record"
            record["msg"] = str(e)
            self.client[self.mongo_db][f"{endpoint}_errors"].insert_one(record)
            self.checkpoint[endpoint].update_record(
                self.mongo_db, endpoint, keys={"_id": identifier})
            print("=== ERROR ===")
            print(e)
            print(identifier)
            return

        record = xmltodict.parse(raw_record)
        record["_id"] = identifier
        try:
            if "error" in record["OAI-PMH"].keys():
                self.client[self.mongo_db][f"{endpoint}_invalid"].insert_one(
                    record)
            else:
                self.client[self.mongo_db][f"{endpoint}_records"].insert_one(
                    record)
            self.checkpoint[endpoint].update_record(
                self.mongo_db, endpoint, keys={"_id": identifier})
        except Exception as e:
            print("=== ERROR: ", e, endpoint, file=sys.stderr)
        # performing "atomic" operation here(to be sure it was inserted)
        finally:
            rcount = self.client[self.mongo_db][f"{endpoint}_records"].count_documents({"_id": identifier})
            icount = self.client[self.mongo_db][f"{endpoint}_invalid"].count_documents({"_id": identifier})
            if rcount != 0 or icount != 0:
                self.checkpoint[endpoint].update_record(
                    self.mongo_db, endpoint, keys={"_id": identifier})

    def process_records(self, client: Client, identifiers: list, metadataPrefix: str, endpoint: str):
        """
        This method makes a loop over the record to perform the request.
        Also reports the progress in stdout every 1000 records.

        Parameters:
        ---------
        client: oaipmh.client
            oaipmh client instance
        identifiers:list
            record ids
        metadataPrefix:str
            metadata type for xml schema ex: dim, xoai, mods, oai_dc (default: oai_dc)
        endpoint:str
            name of the endpoint to process and the MongoDb collection name
        """
        count = 0
        size = len(identifiers)
        for identifier in identifiers:
            self.process_record(
                client, identifier["_id"], metadataPrefix, endpoint)
            if count % 1000 == 0:
                print(
                    f"=== INFO: Downloaded {count} of {size} ({(count/size)*100:.2f}%) for {endpoint}")
            count += 1

    def process_endpoint(self, endpoint: str):
        """
        Method to parse endpoint config, handle checkpoint and process records.

        Parameters:
        ---------
        endpoint:str
            name of the endpoint to process and the MongoDb collection name
        checkpoint:bool
            Bool to enable checkpointing
        """
        url = self.endpoints[endpoint]["url"]
        metadataPrefix = self.endpoints[endpoint]["metadataPrefix"]
        selective = self.endpoints[endpoint]["checkpoint"]["selective"]
        checkpoint = self.endpoints[endpoint]["checkpoint"]["enabled"]
        if selective:
            self.checkpoint[endpoint] = OxomocCheckPointSelective(
                self.mongodb_uri)
        else:
            self.checkpoint[endpoint] = OxomocCheckPoint(self.mongodb_uri)
        if checkpoint:
            self.checkpoint[endpoint].create(
                url, self.mongo_db, endpoint, metadataPrefix)

        print(f"\n=== Processing {endpoint} from {url} ")
        if self.checkpoint[endpoint].exists_records(self.mongo_db, endpoint):
            client = Client(url, force_http_get=self.force_http_get)
            record_ids = self.checkpoint[endpoint].get_records_regs(
                self.mongo_db, endpoint)
            self.process_records(client, record_ids, metadataPrefix, endpoint)
        else:
            print(
                f"=== Error: records checkpoint for {endpoint} not found, create it first with ...")
            print(f"=== Error: Skipping records from {url} in {endpoint}")

    def run(self, jobs: int = None):
        """
        Method to start the harvesting of the data in the multiples endpoints in parallel.
        You have to create the checkpoint first, before call this method.

        Parameters:
        ----------
        jobs:int
            number of jobs for parallel execution, if the value is None, it will
            take the number of threads available in the cpu.
        """
        endpoints_names = self.endpoints.keys()
        if jobs is None:
            jobs = psutil.cpu_count()
        if jobs > len(endpoints_names):
            jobs = len(endpoints_names)
        endpoints = []
        for endpoint in endpoints_names:
            if self.endpoints[endpoint]["enabled"]:
                endpoints.append(endpoint)
            else:
                print(f"=== INFO: repository {endpoint} is disabled, skipped!")
        Parallel(n_jobs=jobs, backend='threading', verbose=10)(delayed(self.process_endpoint)(
            endpoint) for endpoint in endpoints)



from joblib import Parallel, delayed
from oaipmh.client import Client
from pymongo import MongoClient
import psutil


class OxomocCheckPoint:
    """
    Class to handle checkpoints for Colav OAI-PMH
    """

    def __init__(self, mongodb_uri="mongodb://localhost:27017/"):
        """
        CheckPoint constructor

        Parameters:
        ----------
        mongodb_uri:str
            MongoDB connection string uri
        """
        self.client = MongoClient(mongodb_uri)

    def create(self, base_url: str, mongo_db: str, mongo_collection: str, metadataPrefix='oai_dc', force_http_get=True):
        """
        Method to create the checkpoint, this allows to save all the ids for records and sets
        in order to know what was downloaded.
        All the checkpints are saved in the mongo collections

        Parameters:
        ----------
        base_url:str
            D-Space endpoint url
        mongo_db:str
            MongoDB database name
        mongo_collection:str
            MongoDB collection name
        metadataPrefix:str
            metadata type for xml schema ex: dim, xoai, mods, oai_dc (default: oai_dc)
        force_http_get:bool
            force to use get instead post for requests
        """

        client = Client(base_url, force_http_get=force_http_get)
        try:
            print(
                f"=== INFO: Checking identity for {mongo_collection} from  {base_url}")
            identity = client.identify()
        except BaseException as err:
            print(f"=== ERROR: Unexpected {err=}, {type(err)=}")
            print(f"=== ERROR: CheckPoint can not be created for {base_url}")
            return
        if metadataPrefix not in [i[0] for i in client.listMetadataFormats()]:
            print(
                f"=== ERROR: metadataPrefix {metadataPrefix}, not supported for {base_url}")
            print(
                f"=== ERROR: CheckPoint can not be created for {mongo_collection} omitting..")
            return

        print(
            f"=== INFO: Creating CheckPoint for {mongo_collection} from  {base_url}")
        info = {}
        info["repository_name"] = identity.repositoryName()
        info["admin_emails"] = identity.adminEmails()
        info["base_url"] = identity.baseURL()
        info["protocol_version"] = identity.protocolVersion()
        info["earliest_datestamp"] = identity.earliestDatestamp()
        info["granularity"] = identity.granularity()

        self.client[mongo_db][f"{mongo_collection}_identity"].drop()
        self.client[mongo_db][f"{mongo_collection}_identity"].insert_one(info)

        if self.exists_records(mongo_db, mongo_collection):
            print(
                f"=== Getting Records ids from {base_url}  for {mongo_collection}")
            print("=== Calling incremental checkpoints")
            old_ids = list(self.client[mongo_db]
                           [f"{mongo_collection}_identifiers"].find({}, {'_ids': 1}))
            old_ids = [i["_id"] for i in old_ids]
        else:
            old_ids = []
            print(
                f"=== Getting Records ids from {base_url}  for {mongo_collection}")

        ids = client.listIdentifiers(metadataPrefix=metadataPrefix)
        identifiers = []
        counter = 0
        for i in ids:
            _id = i.identifier()
            if _id not in old_ids:
                identifier = {}
                identifier["_id"] = _id
                identifier["datestamp"] = i.datestamp()
                identifier["deleted"] = i.isDeleted()
                identifier["setspec"] = i.setSpec()
                identifier["downloaded"] = False
                identifiers.append(identifier)
            counter += 1
            if counter % 1000 == 0:
                print(
                    f"=== INFO: CheckPoint: Processed {counter} records for {mongo_collection}")
        identifiers = list(
            {item['_id']: item for item in identifiers}.values())
        if len(identifiers) > 0:
            self.client[mongo_db][f"{mongo_collection}_identifiers"].insert_many(
                identifiers)
        print("=== INFO: Records CheckPoint total records found = {} for {}".format(
            len(identifiers), f"{mongo_collection}_identifiers"))

    def exists_records(self, mongo_db: str, mongo_collection: str):
        """
        Method to check if the checkpoints already exists for records.

        Parameters:
        ----------
        mongo_db:str
            MongoDB database name
        mongo_collection:str
            MongoDB collection name
        """
        ckp_rec = f"{mongo_collection}_identifiers"
        collections = self.client[mongo_db].list_collection_names()
        return ckp_rec in collections

    def drop(self, mongo_db: str, mongo_collection: str):
        """
        Method to delete all the checkpoints.

        Parameters:
        ----------
        mongo_db:str
            MongoDB database name
        mongo_collection:str
            MongoDB collection name
        """
        self.client[mongo_db][f"{mongo_collection}_identity"].drop()
        self.client[mongo_db][f"{mongo_collection}_identifiers"].drop()

    def update_record(self, mongo_db: str, mongo_collection: str, keys: dict):
        """
        Method to update the status of a record in the checkpoint

        Parameters:
        ----------
        mongo_db:str
            MongoDB database name
        mongo_collection:str
            MongoDB collection name
        keys:dict
            Dictionary with _id and other required values to perform the update.
        """
        self.client[mongo_db][f"{mongo_collection}_identifiers"].update_one(
            keys, {"$set": {"downloaded": 1}})

    def get_records_regs(self, mongo_db: str, mongo_collection: str):
        """
        Function to get registers from the records ckp collection that are not downloaded

        Parameters:
        ----------
        mongo_db:str
            MongoDB database name
        mongo_collection:str
            MongoDB collection name

        Returns:
        ----------
        list
            ids of records not downloaded.
        """
        ckp_col = self.client[mongo_db][f"{mongo_collection}_identifiers"]
        ckpdata = list(ckp_col.find(
            {"$and": [{"downloaded": False}, {"deleted": False}]}, {"downloaded": 0}))
        return ckpdata

    def run(self, endpoints: dict, mongo_db: str, jobs: int = None):
        """
        Method to create in parallel the checkpoints,
        every thread for endpoint

        Parameters:
        ----------
        endpoints: dict
            dictionary with the endpoints
        mongo_db: str
            database name
        jobs: int
            number of threads for the parallel execution,
            if None maximum allowed by the cpu.
        """
        if jobs is None:
            jobs = psutil.cpu_count()
        Parallel(n_jobs=jobs, backend='threading', verbose=10)(delayed(self.create)(
            endpoints[key]["url"], mongo_db, key, endpoints[key]["metadataPrefix"]) for key in endpoints.keys())
